{ File   : PMS2WAV.dpr
  Author : Deniz Oezmen
  Created: 2007-04-01
  Changed: 2007-08-05

  Jack Orlando - Director's Cut PMS converter.
}
program PMS2WAV;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TPMSHeader = packed record
                 Unknown1  : Longword;
                 DataSize,
                 UncompSize: Longint;
                 Unknown2  : Longword;
                 Sample0,
                 Index0,
                 Sample1,
                 Index1    : Smallint;
               end;

var
  WAVHeader    : packed array[0..43] of Char =
    ( 'R',  'I',  'F',  'F', #$00, #$00, #$00, #$00,  'W',  'A',  'V',  'E',
      'f',  'm',  't',  ' ', #$10, #$00, #$00, #$00, #$01, #$00, #$01, #$00,
     #$22, #$56, #$00, #$00, #$22, #$56, #$00, #$00, #$02, #$00, #$10, #$00,
     'd',  'a',  't',  'a',  #$00, #$00, #$00, #$00);

function OpenPMSFile(var PMSFile: file; PMSName: string): Boolean;
begin
  FileMode := fmOpenRead;
  AssignFile(PMSFile, PMSName);

  {$I-}
  Reset(PMSFile, 1);
  {$I+}
  OpenPMSFile := (IOResult = 0)
end;

function ReadHeader(var PMSFile: file; out PMSHeader: TPMSHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(PMSFile, PMSHeader, SizeOf(PMSHeader), BytesRead);

  Result := (BytesRead = SizeOf(PMSHeader)) and
    (PMSHeader.DataSize + SizeOf(PMSHeader) <= FileSize(PMSFile)) and
    (PMSHeader.DataSize shl 2 = PMSHeader.UncompSize)
end;

procedure IMA_GetNewSample(var Sample: Smallint; var Idx: Byte;
  SampleCode: Byte);
const
  StepTab: array[0..88] of Smallint =
    ($0007, $0008, $0009, $000a, $000b, $000c, $000d, $000e,
     $0010, $0011, $0013, $0015, $0017, $0019, $001c, $001f,
     $0022, $0025, $0029, $002d, $0032, $0037, $003c, $0042,
     $0049, $0050, $0058, $0061, $006b, $0076, $0082, $008f,
     $009d, $00ad, $00be, $00d1, $00e6, $00fd, $0117, $0133,
     $0151, $0173, $0198, $01c1, $01ee, $0220, $0256, $0292,
     $02d4, $031c, $036c, $03c3, $0424, $048e, $0502, $0583,
     $0610, $06ab, $0756, $0812, $08e0, $09c3, $0abd, $0bd0,
     $0cff, $0e4c, $0fba, $114c, $1307, $14ee, $1706, $1954,
     $1bdc, $1ea5, $21b6, $2515, $28ca, $2cdf, $315b, $364b,
     $3bb9, $41b2, $4844, $4f7e, $5771, $602f, $69ce, $7462,
     $7fff);
  IndexTab: array[0..15] of Smallint =
    ( -1, -1, -1, -1,  2,  4,  6,  8,
      -1, -1, -1, -1,  2,  4,  6,  8);
var
  Samp0,
  Idx0,
  SampX,
  Diff : Longint;
begin
  // expand parameters' value range for computations
  Samp0 := Longint(Sample);
  Idx0  := Longint(Idx);
  SampX := Longint(SampleCode);

  // calculate difference
  Diff := 0;
  if SampX and $04 <> 0 then
    Diff := Diff + StepTab[Idx0];
  if SampX and $02 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 1);
  if SampX and $01 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 2);
  Diff := Diff + (StepTab[Idx0] shr 3);
  if SampX and $08 <> 0 then
    Diff := -Diff;

  // calculate new sample
  Samp0 := Samp0 + Diff;
  if Samp0 > 32767 then
    Samp0 := 32767
  else if Samp0 < -32768 then
    Samp0 := -32768;

  // calculate new step table index
  Idx0 := Idx0 + IndexTab[SampX];
  if Idx0 > 88 then
    Idx0 := 88
  else if Idx0 < 0 then
    Idx0 := 0;

  // write values back to parameters
  Sample := Smallint(Samp0);
  Idx := Byte(Idx0)
end;

procedure IMA_ConvertBlock(const InBuf: array of Byte;
  var OutBuf: array of Smallint; const PMSHeader: TPMSHeader);
var
  i, 
  Channel,
  InPos,
  OutPos : Longint;
  Indexes: array[0..1] of Byte;
  Samples: array[0..1] of Smallint;
  Codes  : array[0..1] of Longword;
begin
  with PMSHeader do
  begin
    Indexes[0] := Index0;
    Indexes[1] := Index1;
    Samples[0] := Sample0;
    Samples[1] := Sample1
  end;

  InPos := 0;
  OutPos := 0;

  while InPos < Length(InBuf) do
  begin
    // sample codes for each channel are organized in 4 byte structures
    for Channel := 0 to 1 do
    begin
      Move(InBuf[InPos], Codes[Channel], 4);
      Inc(InPos, 4)
    end;

    // decode samples
    for i := 0 to 7 do
      for Channel := 0 to 1 do
      begin
        IMA_GetNewSample(Samples[Channel], Indexes[Channel],
          Codes[Channel] and $0f);
        Codes[Channel] := Codes[Channel] shr 4;

        OutBuf[OutPos] := Samples[Channel];
        Inc(OutPos)
      end
  end
end;

procedure ConvertPMS(var PMSFile: file; const PMSName: string;
  const PMSHeader: TPMSHeader);
var
  InBuf         : array of Byte;
  OutBuf        : array of Smallint;
  Channels,
  BlockAlign,
  BitsPerSample : Word;
  Frequency,
  BytesPerSecond,
  UncompSize    : Longint;
  WAVFile       : file;
begin
  SetLength(InBuf, PMSHeader.DataSize);
  SetLength(OutBuf, PMSHeader.UncompSize shr 1);

  {$I-}
  BlockRead(PMSFile, InBuf[0], Length(InBuf));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  read error');
    Exit
  end;

  // wave parameters
  Channels := 2;
  Frequency := 44100;
  BytesPerSecond := Frequency * Channels * 2;
  BlockAlign := Channels * 2;
  BitsPerSample := 16;
  UncompSize := PMSHeader.UncompSize;

  // prepare wave header
  Inc(UncompSize, 36);

  Move(UncompSize, WAVHeader[4], 4);
  Move(Channels, WAVHeader[22], 2);
  Move(Frequency, WAVHeader[24], 4);
  Move(BytesPerSecond, WAVHeader[28], 4);
  Move(BlockAlign, WAVHeader[32], 2);
  Move(BitsPerSample, WAVHeader[34], 2);

  Dec(UncompSize, 36);

  Move(UncompSize, WAVHeader[40], 4);

  // open output
  FileMode := fmOpenWrite;
  AssignFile(WAVFile, Format('%s.wav', [PMSName]));

  {$I-}
  Rewrite(WAVFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  write error');
    Exit
  end;

  IMA_ConvertBlock(InBuf, OutBuf, PMSHeader);

  // write wave header and data
  {$I-}
  BlockWrite(WAVFile, WAVHeader[0], SizeOf(WAVHeader));
  BlockWrite(WAVFile, OutBuf[0], Length(OutBuf) shl 1);
  {$I+}
  if IOResult <> 0 then
    WriteLn('  write error');

  CloseFile(WAVFile)
end;

procedure ScanFiles(FileParam: string);
var
  FileName,
  Dir,
  PMSName  : string;
  PMSFile  : file;
  PMSHeader: TPMSHeader;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      PMSName := Dir + FileName;

      WriteLn(PMSName);

      // if this file is accessible, analyze it
      if OpenPMSFile(PMSFile, PMSName) then
      begin
        if ReadHeader(PMSFile, PMSHeader) then
          ConvertPMS(PMSFile, PMSName, PMSHeader)
        else
          WriteLn('  not a valid PMS file');
        Close(PMSFile)
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  FindClose(SearchRec)
end;

begin
  if ParamCount >= 1 then
    ScanFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.
